#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os, sys
import time
import datetime
import io
import urllib3
import httplib2
from PIL import Image, ImageFilter, ImageMath
import numpy as np
import subprocess



# ###########################################################
# Liefert 0 bis zu 35 Tupel aus (Summe, x, y) für das
# in eine 7x5-Matrix zerlegte Bild "arr" als Liste.
# Die Rückgabe-Liste ist absteigend nach Summe sortiert.
def areasByBrightnessV3(arr, cx=7, cy=5, w=91, h=96):
  ret_lst=list()
  for x in range(cx):         # über die Segmente auf der X-Achse
    for y in range(cy):       # über die Segmente auf der Y-Achse
      seg=arr[y*h:y*h+h, 1+x*w:1+x*w+w] # Segment ausschneiden
      br=np.sum(seg)
      if br>2550:             # nur zum Ergebnis zufügen, wenn hell genug
        ret_lst.append((br, x, y))
  f=lambda x:x[0]
  return(sorted(ret_lst, key=f, reverse=True))

# ###########################################################
# Liefert das Bild "img" mit einem Raster zurück.
def insertRaster(img, cx=7, cy=5, w=91, h=96):
  ret_img=np.array(img, copy=True)
  for y in range(1, cy):
    for x in range(cx*w):
      ret_img[y*h, x]=255   # waagerechte Linie
  for x in range(1, cx):
    for y in range(cy*h):
      ret_img[y, x*w]=255   # senkrechte Linie
  return(ret_img)


# ###########################################################
#
class CamImage:
  def __init__(self):
    cam=IP_Cam()
    self.valid=None
    tst=None
    while tst is None:
      self.raw, self.pil=cam.getImage()
      try:
        tst=self.pil.convert('L')
      except:
        print("error - reload image")
        tst=None
    self.u8=np.asarray(tst, np.uint8)
    #self.s16=np.int16(self.u8)

  # ###########################################################
  # Liefert False, wenn das Bild in "self.raw" fehlerhaft ist.
  def isValid(self):
    if self.valid is None:
      t=datetime.datetime.utcnow().strftime("%Y%m%d-%H%M%S-%f")
      fn="/ramdisk/%s__corrupt.jpg"%(t,)
      f=open(fn, "wb")
      f.write(self.raw)
      f.close()
      #if ord(self.raw[-2:-1])!=255 or ord(self.raw[-1:])!=217: # FF D9
      if self.raw[-2:]!=bytes([255, 217]):
        self.valid=False
        print("FF D9")
      else:
        rc=subprocess.call(["identify", "-verbose", fn], 
                            stdout=subprocess.DEVNULL, 
                            stderr=subprocess.DEVNULL)
        self.valid=(rc==0)
        if not self.valid:
          print("identify")
      if(self.valid): # temp. not remove corrupt images
        os.remove(fn)
    return(self.valid)

  # ###########################################################
  # Speichert "self.raw" unter dem Dateinamen "filename".
  def save(self, filename):
    f=open("/ramdisk/%s.jpg"%(filename,), "wb")
    f.write(self.raw)
    f.close()




# ###########################################################
#
class IP_Cam:
  def __init__(self):
    self.CAM_AUTH="camuser:pwd4Cam"
    self.snapshotURL="192.168.178.36/snapshot.cgi"
    self.ctrlURL="192.168.178.36/decoder_control.cgi?command=%d"

    self.http=urllib3.PoolManager()
    self.headers=urllib3.util.make_headers(basic_auth=self.CAM_AUTH)
    self.UP=0
    self.DOWN=2
    self.LEFT=4
    self.RIGHT=6

    self.STOP=1
    self.LEFT_UP=90
    self.RIGHT_UP=91
    self.LEFT_DOWN=92
    self.RIGHT_DOWN=93

  # ###########################################################
  # Schaltet das Blinklicht ein oder aus.
  # Bei "v" 0 oder 1 blinkt die LED schnell oder langsam, bei
  # 2 ist sie aus.
  def setLight(self, v):
    url="192.168.178.36/set_misc.cgi?next_url=light.htm&led_mode=%d"
    r=self.http.request('GET', url%(v), headers=self.headers)

  # ###########################################################
  # Speichert "data" als <timestamp>.jpg unter /ramdisk.
  def saveTmp(self, data):
    t=datetime.datetime.utcnow().strftime("%Y%m%d-%H%M%S-%f")
    f=open("/ramdisk/%s.jpg"%(t,), "wb")
    f.write(data)
    f.close()
    
  # ###########################################################
  # Liefert einen Snapshot als raw und als Pillow-Image.
  def getImage(self):
    r=self.http.request('GET', self.snapshotURL, headers=self.headers)
    tmp=io.BytesIO(r.data)
    #self.saveTmp(tmp.getvalue())
    try:
      image=Image.open(tmp)
    except:
      image=None
    return(tmp.getvalue(), image)

  # ###########################################################
  # Liefert einen Snapshot aus Grauwerten als NumPy-Array.
  def getImageGrayscale(self):
    try:
      img=self.getImage().convert('L')
    except:
      #print("error - retry")
      img=self.getImage().convert('L')
    return(np.asarray(img, np.int8))

  # ###########################################################
  # Bewegt die Kamera "duration" Sekunden lang in die Richtung
  # "direction".
  def gotoDirection(self, direction, duration):
    r=self.http.request('GET', self.ctrlURL%(direction), headers=self.headers)
    if duration<0:
      return
    elif duration>0:
      #print("sleep")
      time.sleep(duration)
    else:
      print("duration=%d"%(duration,))
    r=self.http.request('GET', self.ctrlURL%(1), headers=self.headers)

  # ###########################################################
  # Bewegt die Kamera in Grundstellung.
  def gotoCenter(self):
    r=self.http.request('GET', self.ctrlURL%(25), headers=self.headers)
    while not self.hasStopped():
      time.sleep(0.5)

  # ###########################################################
  # Speichert das Bild "imgAsArray" unter dem Dateinamen
  # "filename".
  def saveImage(self, imgAsArray, filename, raw=False):
    if raw:
      img=imgAsArray
    else:
      img=Image.fromarray(imgAsArray)
    img.save("/ramdisk/%s.jpg"%(filename,))

  # ###########################################################
  # Liefert für ein Grau-Bild in "imgAsArray" eine Liste mit
  # 64 Werten als Pattern. Die Werte entsprechen dem Mittelwert
  # des jeweiligen Bildausschnittes.
  def getImagePattern(self, imgAsArray):
    pat=np.empty([8, 8], dtype=np.int8) # Array anlegen
    arrsh=np.hsplit(imgAsArray, 8)      # Bild horizontal splitten
    for h in range(len(arrsh)):         # über 8 Teilbilder
      arrsv=np.vsplit(arrsh[h], 8)      # Teilbild vertikal splitten
      for v in range(len(arrsv)):       # über 8 Teilbilder
        #self.saveImage(arrsv[v], "b%d%d"%(h,v))
        pat[v, h]=np.mean(arrsv[v])     # Bild-Mittelwert ins Array[y,x]
    return(pat)                         # Array als Pattern liefern

  # ###########################################################
  # Vergleicht die zwei Pattern "pat1" und "pat2" und liefert
  # True, wenn der mittlerer Grauwert jedes Bildausschnitts
  # in beiden Pattern maximal um "maxdelta" voneinander
  # abweicht.
  def compareImagePattern(self, pat1, pat2, maxdelta=2):
    return(np.max(np.absolute(pat1-pat2))<=maxdelta)

  # ###########################################################
  # Vergleicht die zwei Pattern "pat1" und "pat2" und liefert
  # ein Tupel aus Summe der Abweichungen und höchter
  # Abweichung in einem Bildausschnitt.
  def compareImagePatternDetail(self, pat1, pat2):
    tmp=np.absolute(pat1-pat2)
    return(np.sum(tmp), np.max(tmp))

  # ###########################################################
  # Liefert True, wenn zwei Bilder, die 100ms nacheinander
  # von der Kamera gelesen wurden, annähernd gleich sind.
  def hasStopped(self):
    p1=self.getImagePattern(self.getImageGrayscale())
    time.sleep(0.1)
    p2=self.getImagePattern(self.getImageGrayscale())
    return(self.compareImagePattern(p1, p2))

  # ###########################################################
  #
  def gotoMatrixElement(self, x, y):
    #  0,0  1,0  2,0  3,0  4,0  5,0  6,0
    #  0,1  1,1  2,1  3,1  4,1  5,1  6,1
    #  0,2  1,2  2,2 (3,2) 4,2  5,2  6,2
    #  0,3  1,3  2,3  3,3  4,3  5,3  6,3
    #  0,4  1,4  2,4  3,4  4,4  5,4  6,4

    if   y<2: dv=self.UP
    elif y>2: dv=self.DOWN
    else:     dv=None

    if   x<3: dh=self.LEFT
    elif x>3: dh=self.RIGHT
    else:     dh=None

    tv=0.0
    if not dv is None:
      if   y in (0, 4): tv=0.581  # 1.5 Elemente hoch/runter
      elif y in (1, 3): tv=0.325  # 0.5 Elemente hoch/runter

    th=0.0
    if not dh is None:
      if   x in (0, 6): th=2.63   # 2.5 Elemente links/rechts
      elif x in (1, 5): th=1.72   # 1.5 Elemente links/rechts
      elif x in (2, 4): th=0.93   # 0.5 Elemente links/rechts

    if DEBUG: 
      nam={0:"   UP", 2:" DOWN", 4:" LEFT", 6:"RIGHT", None:" none"}
      print("goto(%d,%d)  %s=%5.3f  %s=%4.2f"%(x, y, nam[dv], tv, nam[dh], th))

    if not dv is None and not dh is None:
      # wenn beide Achsen bewegt werden sollen
      if dv==self.UP:
        if dh==self.LEFT:
          if tv>th: # mehr vertikal
            self.gotoDirection(self.LEFT_UP, th)
            self.gotoDirection(self.UP, tv-th)
          else:     # mehr horizontal
            self.gotoDirection(self.LEFT_UP, tv)
            self.gotoDirection(self.LEFT, th-tv)
        else: # dh==self.RIGHT
          if tv>th: # mehr vertikal
            self.gotoDirection(self.RIGHT_UP, th)
            self.gotoDirection(self.UP, tv-th)
          else:     # mehr horizontal
            self.gotoDirection(self.RIGHT_UP, tv)
            self.gotoDirection(self.RIGHT, th-tv)
      else: # dv==self.DOWN
        if dh==self.LEFT:
          if tv>th: # mehr vertikal
            self.gotoDirection(self.LEFT_DOWN, th)
            self.gotoDirection(self.DOWN, tv-th)
          else:     # mehr horizontal
            self.gotoDirection(self.LEFT_DOWN, tv)
            self.gotoDirection(self.LEFT, th-tv)
        else: # dh==self.RIGHT
          if tv>th: # mehr vertikal
            self.gotoDirection(self.RIGHT_DOWN, th)
            self.gotoDirection(self.DOWN, tv-th)
          else:     # mehr horizontal
            self.gotoDirection(self.RIGHT_DOWN, tv)
            self.gotoDirection(self.RIGHT, th-tv)
    else:
      if not dv is None:
        self.gotoDirection(dv, tv)
      if not dh is None:
        self.gotoDirection(dh, th)

# ###########################################################
# (Quick&Dirty) Terminal-Ausgabe eines Hexdumps von dem
# Bytes-Array "bytes_src" mit maximaler Länge "maxlen".
# Bei Übergabe von "maxlen" mit einem negativen Wert, wird
# der gesamte Inhalt von "bytes_src" ausgegeben.
def hexdump(bytes_src, maxlen=1024):
  a=h=""
  cnt=addr=0
  if maxlen<0:  maxlen=len(bytes_src)
  for ci in range(0, min(len(bytes_src), maxlen)):
    c=bytes_src[ci]
    h+="%02x "%(c,)
    if 31<c<127:  a+=chr(c)
    else:         a+="."
    cnt+=1
    if cnt>15:
      print("%06x   %-48s   %s"%(addr, h, a))
      addr+=16
      cnt=0
      a=h=""
  if cnt>0:
    print("%06x   %-48s   %s"%(addr, h, a))

# ###########################################################
# Liefert ein Histogramm von dem Bild "img" als Array mit
# vier Elementen. Vorne die Anzahl der hellen Pixel, hinten
# die Anzahl der dunklen.
def histogram(img):
  h, b=np.histogram(img, [0, 60, 120, 180, 255])
  return(h[::-1])

# ###########################################################
# Liefert die absolute Differnz der NumPy-uint8-Arrays
# "img1" und "img2" als NumPy-uint8-Array.
def differenceImageV6(img1, img2):
  a=img1-img2
  b=np.uint8(img1<img2)*254+1
  return(a*b)

# ###########################################################
# Liefert die Zahl "int_val" als String mit
# Tausender-Trennpunkten.
def intWithDots(int_val):
  rs="{:,}".format(int_val)
  return(rs.replace(",", "."))


# ###########################################################
# Main
if __name__=="__main__":
  DEBUG=False
  nervous=False
  cam=IP_Cam()

  img0=img1=CamImage()

  while True:
    motion_detected=False # True, wenn Bewegung erkannt wurde
    movement_enable=True  # False, wenn nicht neu Positioniert werden soll
    get_new_image=False   # True, um das aktuelle Bild zu verwerfen
    time.sleep(0.1)
    img2=CamImage()

    diff02=differenceImageV6(img0.u8, img2.u8)
    h02=histogram(diff02)   # zur Bestimmung, ob im Bild genug Bewegung vorkommt
    
    if h02[0]>20 or h02[1]>100:
      # Wenn das Differenzbild mindestens 100 Pixel oberhalb von 120
      # oder mindestens 20 Pixel oberhalb von 180 enthält, dann hat
      # möglicherweise Bewegung stattgefunden.

      if img1.isValid() and img2.isValid():
        # Wenn keins der beiden Bilder defekt empfangen wurde, dann
        # könnte das tatsächlich Bewegung sein.
        motion_detected=True
      else:
        get_new_image=True
        if DEBUG: print("fehlerhaft empfangenes Bild")

    if motion_detected:
      diff01=differenceImageV6(img0.u8, img1.u8)
      diff01b=diff01>100
      diff02b=diff02>100
      d1and2=np.logical_and(diff01b, diff02b)
      d2xor=np.logical_xor(diff02b, d1and2)

    if motion_detected:
      t=datetime.datetime.utcnow().strftime("%Y%m%d-%H%M%S-%f")
      if DEBUG:
        cam.saveImage(img0.u8, "%s-0"%(t,))
        cam.saveImage(img1.u8, "%s-1"%(t,))
        cam.saveImage(img2.u8, "%s-2"%(t,))
        cam.saveImage(insertRaster(diff01), "%s-3"%(t,))
        cam.saveImage(insertRaster(diff02), "%s-4"%(t,))
        cam.saveImage(insertRaster(np.uint8(-d1and2-1)), "%s-5"%(t,))
        cam.saveImage(insertRaster(np.uint8(-d2xor-1)), "%s-6"%(t,))
        #cam.saveImage(img2.pil, "%s-4"%(t,), raw=True)
        print("\ndiff02.hist=%s"%(h02,), t)
      else:
        img2.save("motion-%s"%(t,))
    
      # Bild in eine 7x5-Matrix zerlegen
      abb=areasByBrightnessV3(np.uint8(-d2xor-1))
      if len(abb)==0: # doch nicht genug Bewegung (an einem Ort)
        motion_detected=False

    if motion_detected:      
      if DEBUG: print("len(abb)=", len(abb), abb)
      hist, x, y=abb[0]

      if not nervous:
        if x in (2, 3, 4) and y in (1, 2, 3):
          # wenn neue Position in der Bildmitte liegt -> nicht neu positionieren
          movement_enable=False
          if DEBUG: print("bereits mittig")

      if movement_enable:
        if len(abb)>28: # mehr als 28 Elemente über Grundrauschen
          # Zu viele Bereiche über Grundrauschen bedeutet üblicherweise:
          # - ein Bild ist entweder verwackelt oder 
          # - die Kamera hat die Helligkeit angepasst.
          # Das Differenzbild ist jedenfalls unbrauchbar.
          movement_enable=False
          get_new_image=True
          if DEBUG: print("verwackelt", "*"*40)

    if motion_detected and movement_enable:
      #cam.setLight(0)
      cam.gotoMatrixElement(x, y) # neue Position anfahren
      time.sleep(0.2)             # auf Stillstand der Kamera warten
      #cam.setLight(2)
      get_new_image=True

    if get_new_image:
      img1=img2=CamImage()             # Bild von der neuen Position holen
    img0=img1
    img1=img2

